@@ -32,8 +32,8 @@
  * files in the program, then also delete it here.
  */
 /* copyright --> */
-#ifndef _D_SOCKET_CORE_H_
-#define _D_SOCKET_CORE_H_
+#ifndef D_SOCKET_CORE_H
+#define D_SOCKET_CORE_H
 
 #include "common.h"
 
@@ -41,19 +41,9 @@
 #include <cstdlib>
 #include <utility>
 #include <vector>
+#include <memory>
 
 #include "a2netcompat.h"
-
-#ifdef HAVE_LIBSSL
-// for SSL
-# include <openssl/ssl.h>
-# include <openssl/err.h>
-#endif // HAVE_LIBSSL
-#ifdef HAVE_LIBGNUTLS
-# include <gnutls/gnutls.h>
-#endif // HAVE_LIBGNUTLS
-
-#include "SharedHandle.h"
 #include "a2io.h"
 #include "a2netcompat.h"
 #include "a2time.h"
@@ -62,12 +52,18 @@ namespace aria2 {
 
 #ifdef ENABLE_SSL
 class TLSContext;
+class TLSSession;
 #endif // ENABLE_SSL
 
+#ifdef HAVE_LIBSSH2
+class SSHSession;
+#endif // HAVE_LIBSSH2
+
 class SocketCore {
   friend bool operator==(const SocketCore& s1, const SocketCore& s2);
   friend bool operator!=(const SocketCore& s1, const SocketCore& s2);
   friend bool operator<(const SocketCore& s1, const SocketCore& s2);
+
 private:
   // socket type defined in <sys/socket.h>
   int sockType_;
@@ -75,8 +71,13 @@ private:
   sock_t sockfd_;
 
   static int protocolFamily_;
+  static int ipDscp_;
 
-  static std::vector<std::pair<struct sockaddr_storage, socklen_t> > bindAddrs_;
+  static std::vector<SockAddr> bindAddrs_;
+  static std::vector<std::vector<SockAddr>> bindAddrsList_;
+  static std::vector<std::vector<SockAddr>>::iterator bindAddrsListIt_;
+
+  static int socketRecvBufferSize_;
 
   bool blocking_;
   int secure_;
@@ -85,28 +86,27 @@ private:
   bool wantWrite_;
 
 #if ENABLE_SSL
-  static SharedHandle<TLSContext> tlsContext_;
+  // TLS context for client side
+  static std::shared_ptr<TLSContext> clTlsContext_;
+  // TLS context for server side
+  static std::shared_ptr<TLSContext> svTlsContext_;
+
+  std::shared_ptr<TLSSession> tlsSession_;
+
+  /**
+   * Makes this socket secure. The connection must be established
+   * before calling this method.
+   *
+   * If you are going to verify peer's certificate, hostname must be supplied.
+   */
+  bool tlsHandshake(TLSContext* tlsctx, const std::string& hostname);
 #endif // ENABLE_SSL
 
-#ifdef HAVE_LIBSSL
-  // for SSL
-  SSL* ssl;
+#ifdef HAVE_LIBSSH2
+  std::unique_ptr<SSHSession> sshSession_;
 
-  int sslHandleEAGAIN(int ret);
-#endif // HAVE_LIBSSL
-#ifdef HAVE_LIBGNUTLS
-  gnutls_session_t sslSession_;
-  char* peekBuf_;
-  size_t peekBufLength_;
-  size_t peekBufMax_;
-
-  size_t shiftPeekData(char* data, size_t len);
-  void addPeekData(char* data, size_t len);
-  ssize_t gnutlsRecv(char* data, size_t len);
-  ssize_t gnutlsPeek(char* data, size_t len);
-
-  void gnutlsRecordCheckDirection();
-#endif // HAVE_LIBGNUTLS
+  void sshCheckDirection();
+#endif // HAVE_LIBSSH2
 
   void init();
 
@@ -114,14 +114,18 @@ private:
 
   void setSockOpt(int level, int optname, void* optval, socklen_t optlen);
 
-  SocketCore(sock_t sockfd, int sockType);
 public:
   SocketCore(int sockType = SOCK_STREAM);
+
+  // Formally, private constructor, but made public to use with
+  // std::make_shared.
+  SocketCore(sock_t sockfd, int sockType);
+
   ~SocketCore();
 
   sock_t getSockfd() const { return sockfd_; }
 
-  bool isOpen() const { return sockfd_ != (sock_t) -1; }
+  bool isOpen() const { return sockfd_ != (sock_t)-1; }
 
   void setMulticastInterface(const std::string& localAddr);
 
@@ -129,10 +133,20 @@ public:
 
   void setMulticastLoop(unsigned char loop);
 
-  void joinMulticastGroup
-  (const std::string& multicastAddr, uint16_t multicastPort,
-   const std::string& localAddr);
-  
+  void joinMulticastGroup(const std::string& multicastAddr,
+                          uint16_t multicastPort, const std::string& localAddr);
+
+  // Enables TCP_NODELAY socket option if f == true.
+  void setTcpNodelay(bool f);
+
+  // Set DSCP byte
+  void applyIpDscp();
+  static void setIpDscp(int ipDscp)
+  {
+    // Here we prepare DSCP value for IPTOS option, which sets whole DS field
+    ipDscp_ = ipDscp << 2;
+  }
+
   void create(int family, int protocol = 0);
 
   void bindWithFamily(uint16_t port, int family, int flags = AI_PASSIVE);
@@ -140,13 +154,13 @@ public:
   /**
    * Creates a socket and bind it with locahost's address and port.
    * flags is set to struct addrinfo's ai_flags.
-   * @param port port to listen. If 0 is specified, os automaticaly
-   * choose avaiable port.
+   * @param port port to listen. If 0 is specified, os automatically
+   * choose available port.
    */
   void bind(uint16_t port, int flags = AI_PASSIVE);
 
-  void bind
-  (const std::string& addr, uint16_t port, int family, int flags = AI_PASSIVE);
+  void bind(const char* addrp, uint16_t port, int family,
+            int flags = AI_PASSIVE);
 
   /**
    * Listens form connection on it.
@@ -155,10 +169,9 @@ public:
   void beginListen();
 
   /**
-   * Stores host address and port of this socket to addrinfo.
-   * @param addrinfo placeholder to store host address and port.
+   * Returns host address, family and port of this socket.
    */
-  void getAddrInfo(std::pair<std::string, uint16_t>& addrinfo) const;
+  Endpoint getAddrInfo() const;
 
   /**
    * Stores address of this socket to sockaddr.  len must be
@@ -166,8 +179,7 @@ public:
    * stored in sockaddr and actual size of address structure is stored
    * in len.
    */
-  void getAddrInfo
-  (struct sockaddr_storage& sockaddr, socklen_t& len) const;
+  void getAddrInfo(sockaddr_union& sockaddr, socklen_t& len) const;
 
   /**
    * Returns address family of this socket.
@@ -176,17 +188,16 @@ public:
   int getAddressFamily() const;
 
   /**
-   * Stores peer's address and port to peerinfo.
-   * @param peerinfo placeholder to store peer's address and port.
+   * Returns peer's address, family and port.
    */
-  void getPeerInfo(std::pair<std::string, uint16_t>& peerinfo) const;
+  Endpoint getPeerInfo() const;
 
   /**
    * Accepts incoming connection on this socket.
    * You must call beginListen() before calling this method.
-   * @return accepted socket. The caller must delete it after using it.
+   * @return accepted socket.
    */
-  SocketCore* acceptConnection() const;
+  std::shared_ptr<SocketCore> acceptConnection() const;
 
   /**
    * Connects to the server named host and the destination port is port.
@@ -195,8 +206,10 @@ public:
    * the connection is established.
    * @param host hostname or ip address to connect to
    * @param port service port number to connect to
+   * @param tcpNodelay true to disable Nagle algorithm
    */
-  void establishConnection(const std::string& host, uint16_t port);
+  void establishConnection(const std::string& host, uint16_t port,
+                           bool tcpNodelay = true);
 
   void setNonBlockingMode();
 
@@ -240,25 +253,16 @@ public:
    * @param data data to write
    * @param len length of data
    */
-  ssize_t writeData(const char* data, size_t len);
+  ssize_t writeData(const void* data, size_t len);
   ssize_t writeData(const std::string& msg)
   {
     return writeData(msg.c_str(), msg.size());
   }
-  ssize_t writeData(const unsigned char* data, size_t len)
-  {
-    return writeData(reinterpret_cast<const char*>(data), len);
-  }
 
-  ssize_t writeData(const char* data, size_t len,
-                    const std::string& host, uint16_t port);
+  ssize_t writeData(const void* data, size_t len, const std::string& host,
+                    uint16_t port);
 
-  ssize_t writeData(const unsigned char* data, size_t len,
-                    const std::string& host,
-                    uint16_t port)
-  {
-    return writeData(reinterpret_cast<const char*>(data), len, host, port);
-  }
+  ssize_t writeVector(a2iovec* iov, size_t iovcnt);
 
   /**
    * Reads up to len bytes from this socket.
@@ -276,62 +280,47 @@ public:
    * @param len the maximum size data can store. This method assigns
    * the number of bytes read to len.
    */
-  void readData(char* data, size_t& len);
+  void readData(void* data, size_t& len);
 
-  void readData(unsigned char* data, size_t& len)
-  {
-    readData(reinterpret_cast<char*>(data), len);
-  }
+  // sender.addr will be numerihost assigned.
+  ssize_t readDataFrom(void* data, size_t len, Endpoint& sender);
 
-  ssize_t readDataFrom(char* data, size_t len,
-                       std::pair<std::string /* numerichost */,
-                       uint16_t /* port */>& sender);
+#ifdef ENABLE_SSL
+  // Performs TLS server side handshake. If handshake is completed,
+  // returns true. If handshake has not been done yet, returns false.
+  bool tlsAccept();
 
-  ssize_t readDataFrom(unsigned char* data, size_t len,
-                       std::pair<std::string /* numerichost */,
-                       uint16_t /* port */>& sender)
-  {
-    return readDataFrom(reinterpret_cast<char*>(data), len, sender);
-  }
+  // Performs TLS client side handshake. If handshake is completed,
+  // returns true. If handshake has not been done yet, returns false.
+  //
+  // If you are going to verify peer's certificate, hostname must be
+  // supplied.
+  bool tlsConnect(const std::string& hostname);
+#endif // ENABLE_SSL
 
-  /**
-   * Reads up to len bytes from this socket, but bytes are not removed from
-   * this socket.
-   * This method internally calls isReadable(). The parameter timeout is used
-   * for this method call.
-   * @param data holder to store data.
-   * @param len the maximum size data can store. This method assigns
-   * the number of bytes read to len.
-   */
-  void peekData(char* data, size_t& len);
+#ifdef HAVE_LIBSSH2
+  // Performs SSH handshake
+  bool sshHandshake(const std::string& hashType, const std::string& digest);
+  // Performs SSH authentication using username and password.
+  bool sshAuthPassword(const std::string& user, const std::string& password);
+  // Starts sftp session and open remote file |path|.
+  bool sshSFTPOpen(const std::string& path);
+  // Closes sftp remote file gracefully
+  bool sshSFTPClose();
+  // Gets total length and modified time for remote file currently
+  // opened.  |path| is used for logging.
+  bool sshSFTPStat(int64_t& totalLength, time_t& mtime,
+                   const std::string& path);
+  // Seeks file position to |pos|.
+  void sshSFTPSeek(int64_t pos);
+  bool sshGracefulShutdown();
+#endif // HAVE_LIBSSH2
 
-  void peekData(unsigned char* data, size_t& len)
-  {
-    peekData(reinterpret_cast<char*>(data), len);
-  }
+  bool operator==(const SocketCore& s) { return sockfd_ == s.sockfd_; }
 
-  /**
-   * Makes this socket secure.
-   * If the system has not OpenSSL, then this method do nothing.
-   * connection must be established  before calling this method.
-   *
-   * If you are going to verify peer's certificate, hostname must be supplied.
-   */
-  bool initiateSecureConnection(const std::string& hostname="");
+  bool operator!=(const SocketCore& s) { return !(*this == s); }
 
-  void prepareSecureConnection();
-
-  bool operator==(const SocketCore& s) {
-    return sockfd_ == s.sockfd_;
-  }
-
-  bool operator!=(const SocketCore& s) {
-    return !(*this == s);
-  }
-
-  bool operator<(const SocketCore& s) {
-    return sockfd_ < s.sockfd_;
-  }
+  bool operator<(const SocketCore& s) { return sockfd_ < s.sockfd_; }
 
   std::string getSocketError() const;
 
@@ -348,8 +337,16 @@ public:
    */
   bool wantWrite() const;
 
+  // Returns buffered data which are already received.  This data was
+  // already read from socket, and ready to read without reading
+  // socket.
+  size_t getRecvBufferedLength() const;
+
 #ifdef ENABLE_SSL
-  static void setTLSContext(const SharedHandle<TLSContext>& tlsContext);
+  static void
+  setClientTLSContext(const std::shared_ptr<TLSContext>& tlsContext);
+  static void
+  setServerTLSContext(const std::shared_ptr<TLSContext>& tlsContext);
 #endif // ENABLE_SSL
 
   static void setProtocolFamily(int protocolFamily)
@@ -357,6 +354,9 @@ public:
     protocolFamily_ = protocolFamily;
   }
 
+  static void setSocketRecvBufferSize(int size);
+  static int getSocketRecvBufferSize();
+
   // Bind socket to interface. interface may be specified as a
   // hostname, IP address or interface name like eth0.  If the given
   // interface is not found or binding socket is failed, exception
@@ -366,10 +366,16 @@ public:
   // We cannot use interface as an argument because it is a reserved
   // keyword in MSVC.
   static void bindAddress(const std::string& iface);
+  static void bindAllAddress(const std::string& ifaces);
 
-  friend void getInterfaceAddress
-  (std::vector<std::pair<struct sockaddr_storage, socklen_t> >& ifAddrs,
-   const std::string& iface, int family, int aiFlags);
+  // Collects IP addresses of given interface iface and stores in
+  // ifAddres. iface may be specified as a hostname, IP address or
+  // interface name like eth0. You can limit the family of IP
+  // addresses to collect using family argument. aiFlags is passed to
+  // getaddrinfo() as hints.ai_flags. No throw.
+  static std::vector<SockAddr> getInterfaceAddress(const std::string& iface,
+                                                   int family = AF_UNSPEC,
+                                                   int aiFlags = 0);
 };
 
 // Set default ai_flags. hints.ai_flags is initialized with this
@@ -380,19 +386,47 @@ void setDefaultAIFlags(int flags);
 // flags|DEFAULT_AI_FLAGS is used as ai_flags.  You can override
 // DEFAULT_AI_FLAGS value by calling setDefaultAIFlags() with new
 // flags.
-int callGetaddrinfo
-(struct addrinfo** resPtr, const char* host, const char* service, int family,
- int sockType, int flags, int protocol);
+int callGetaddrinfo(struct addrinfo** resPtr, const char* host,
+                    const char* service, int family, int sockType, int flags,
+                    int protocol);
 
-// Collects IP addresses of given inteface iface and stores in
-// ifAddres. iface may be specified as a hostname, IP address or
-// interface name like eth0. You can limit the family of IP addresses
-// to collect using family argument. aiFlags is passed to
-// getaddrinfo() as hints.ai_flags. No throw.
-void getInterfaceAddress
-(std::vector<std::pair<struct sockaddr_storage, socklen_t> >& ifAddrs,
- const std::string& iface, int family = AF_UNSPEC, int aiFlags = 0);
+// Provides functionality of inet_ntop using getnameinfo.  The return
+// value is the exact value of getnameinfo returns. You can get error
+// message using gai_strerror(3).
+int inetNtop(int af, const void* src, char* dst, socklen_t size);
+
+// Provides functionality of inet_pton using getBinAddr.  If af is
+// AF_INET, dst is assumed to be the pointer to struct in_addr.  If af
+// is AF_INET6, dst is assumed to be the pointer to struct in6_addr.
+//
+// This function returns 0 if it succeeds, or -1.
+int inetPton(int af, const char* src, void* dst);
+
+namespace net {
+
+// Stores binary representation of IP address ip which is represented
+// in text.  ip must be numeric IPv4 or IPv6 address. dest must be
+// allocated by caller before the call. For IPv4 address, dest must be
+// at least 4. For IPv6 address, dest must be at least 16. Returns the
+// number of bytes written in dest, that is 4 for IPv4 and 16 for
+// IPv6. Return 0 if error occurred.
+size_t getBinAddr(void* dest, const std::string& ip);
+
+// Verifies hostname against presented identifiers in the certificate.
+// The implementation is based on the procedure described in RFC 6125.
+bool verifyHostname(const std::string& hostname,
+                    const std::vector<std::string>& dnsNames,
+                    const std::vector<std::string>& ipAddrs,
+                    const std::string& commonName);
+// Checks public IP address are configured for each family: IPv4 and
+// IPv6. The result can be obtained using getIpv4AddrConfigured() and
+// getIpv6AddrConfigured() respectively.
+void checkAddrconfig();
+bool getIPv4AddrConfigured();
+bool getIPv6AddrConfigured();
+
+} // namespace net
 
 } // namespace aria2
 
-#endif // _D_SOCKET_CORE_H_
+#endif // D_SOCKET_CORE_H
